/*
 * Copyright 2012 Steven Gribble
 *
 *  This file is part of the UW CSE333 project sequence (333proj12).
 *
 *  333proj12 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  333proj12 is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj12.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdexcept>

#include "./JSONParser.h"

// A function for comparing two ObjectTreeStruct's.
int ObjectTreeComparatorFn(void *a, void *b);

// A function for freeing an ObjectTreeStruct.
void ObjectTreeFreeFn(void *a);

// A function for freeing a linked list item.
void LinkedListFreeFn(void *a);

// Utility functions for the parser.
void RecursiveParse(std::string &str, JSONValue **next);
void ParseObject(std::string &pr, JSONObject **object);
void ParseArray(std::string &pr, JSONArray **array);
void ParseString(std::string &pr, JSONString **string);
void ParseNumber(std::string &pr, JSONNumber **number);
void ParseTrue(std::string &pr, JSONBoolean **t);
void ParseFalse(std::string &pr, JSONBoolean **f);
void ParseNull(std::string &pr, JSONNull **n);

//
// JSONValue-related public functions.
//

JSONValue::~JSONValue() { }

JSONValue *JSONValue::JSONParse(const std::string &data) {
  JSONValue *p = NULL;
  std::string localstr = data;

  RecursiveParse(localstr, &p);
  return p;
}

std::ostream& operator<<(std::ostream &o, const JSONValue &b) {
  return b.Dump(o);
}

//
// JSONObject-related public functions.
//

JSONObject::JSONObject() {
  tree_ = AllocateAVLTree();
  if (tree_ == NULL)
    throw std::runtime_error("Out of memory allocating the AVL "
                             "tree in JSONObject()");
}

JSONObject::~JSONObject() {
  FreeAVLTree(tree_, &ObjectTreeFreeFn);
}

JSONValue *JSONObject::Get(const std::string &key) {
  // Try to look up the "key" in the tree; fail if not found.
  JSONObject::ObjectTreeStruct lookup_st = { key, NULL };
  JSONObject::ObjectTreeStructPtr payload = NULL;
  if (LookupInAVLTree(tree_,
                      reinterpret_cast<void *>(&lookup_st),
                      reinterpret_cast<void **>(&payload),
                      &ObjectTreeComparatorFn) == 0)
    return NULL;

  // Found the key, so return the payload's value.
  return payload->value;
}

bool JSONObject::Put(const std::string &key, JSONValue *value_to_add) {
  // Allocate space for the new payload.
  ObjectTreeStructPtr payload = new ObjectTreeStruct;
  if (payload == NULL)
    return false;

  // Copy in the key and payload.
  payload->key = key;
  payload->value = reinterpret_cast<JSONValue *>(value_to_add);

  // Do the insert.
  if (InsertIntoAVLTree(tree_,
                        payload,
                        &ObjectTreeComparatorFn) == 0) {
    delete payload;
    return false;
  }

  // Success!
  return true;
}

std::ostream& JSONObject::Dump(std::ostream &o) const {
  o << "{ ";

  // Iterate through the tree, putting in the key / val pairs
  AVLTreeIterPtr tip = AVLTreeMakeIterator(tree_, INORDER);
  uint64_t numels = NumElementsInAVLTree(tree_);
  for (uint64_t i = 0U; i < numels; i++) {
    ObjectTreeStructPtr otsp = NULL;

    AVLTreeIteratorGetPayload(tip, reinterpret_cast<void **>(&otsp));
    o << "\"" << otsp->key << "\": " << *(otsp->value);
    if (i != numels-1)
      o << ", ";

    AVLTreeIteratorNext(tip);
  }
  if (tip != NULL)
    AVLTreeIteratorFree(tip);

  o << " }";
  return o;
}

//
// JSONArray-related public functions
//

JSONArray::JSONArray() {
  list_ = AllocateLinkedList();
  if (list_ == NULL)
    throw std::runtime_error("Out of memory allocating the "
                             "list in JSONArray()");
}

JSONArray::~JSONArray() {
  FreeLinkedList(list_, &LinkedListFreeFn);
}

JSONValue *JSONArray::operator[](const uint64_t index) {
  JSONValue *payload = NULL;
  // STEP 3: implement operator[]().

  return payload;
}

bool JSONArray::Append(JSONValue *value_to_append) {
  // Do the append.
  if (!AppendLinkedList(list_, value_to_append))
    return false;

  // Success!
  return true;
}

uint64_t JSONArray::Length() {
  return NumElementsInLinkedList(list_);
}

std::ostream& JSONArray::Dump(std::ostream &o) const {
  // STEP 2.  Implement JSONArray::Dump().

  return o;
}

//
// JSONString-related public functions
//

std::ostream& JSONString::Dump(std::ostream &o) const {
  // XXX -- have to encode, add " ", dump that.
  std::string encoded_string = JSONEscapeString((unsigned char *) str_.c_str());

  o << "\"" << encoded_string << "\"";
  return o;
}

//
// JSONNumber-related functions
//

std::ostream& JSONNumber::Dump(std::ostream &o) const {
  return o << num_;
}

//
// JSONBoolean-related functions
//

std::ostream& JSONBoolean::Dump(std::ostream &o) const {
  return o << (val_ ? "true" : "false");
}

//
// JSONNull-related functions
//

std::ostream& JSONNull::Dump(std::ostream &o) const {
  return o << "null";
}

//
// Private functions
//

int ObjectTreeComparatorFn(void *a, void *b) {
  // Cast to the correct type.
  JSONObject::ObjectTreeStructPtr otsp_a = (JSONObject::ObjectTreeStructPtr) a;
  JSONObject::ObjectTreeStructPtr otsp_b = (JSONObject::ObjectTreeStructPtr) b;

  // Defensive programming.
  assert(otsp_a != NULL);
  assert(otsp_b != NULL);

  // Do the string comparison.
  int res = strcmp(otsp_a->key.c_str(), otsp_b->key.c_str());
  if (res < 0)
    return -1;
  if (res > 0)
    return 1;
  return 0;
}

void ObjectTreeFreeFn(void *a) {
  // Cast to the correct type.
  JSONObject::ObjectTreeStructPtr otsp_a = (JSONObject::ObjectTreeStructPtr) a;
  assert(otsp_a != NULL);

  // Free the value.
  delete otsp_a->value;

  // Free the payload structure.
  delete otsp_a;
}

void LinkedListFreeFn(void *a) {
  // Delete the value.
  JSONValue *x = reinterpret_cast<JSONValue *>(a);
  delete x;
}

void RecursiveParse(std::string &str, JSONValue **next) {
  // Trim any leading or trailing whitespace.
  boost::trim(str);

  // What kind of object is next?  Figure it out and parse it.
  switch (str[0]) {
  case '{':
    ParseObject(str, reinterpret_cast<JSONObject **>(next));
    break;
  case '[':
    ParseArray(str, reinterpret_cast<JSONArray **>(next));
    break;
  case '\"':
    ParseString(str, reinterpret_cast<JSONString **>(next));
    break;
  case '-':
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    ParseNumber(str, reinterpret_cast<JSONNumber **>(next));
    break;
  case 't':
    ParseTrue(str, reinterpret_cast<JSONBoolean **>(next));
    break;
  case 'f':
    ParseFalse(str, reinterpret_cast<JSONBoolean **>(next));
    break;
  case 'n':
    ParseNull(str, reinterpret_cast<JSONNull **>(next));
    break;
  default:
    {
      std::string pe = "parse error: '";
      pe += str[0];
      pe += "'";
      throw std::runtime_error(pe);
    }
    break;
  }
}

void ParseObject(std::string &pr, JSONObject **object) {
  // Skip the leading '{ '
  pr.erase(0, 1);

  // Parse the object name/val pairs
  *object = new JSONObject();
  if (*object == NULL)
    throw std::runtime_error("out of memory allocating JSONObject");

  try {
    while (1) {
      // Trim leading/trailing whitespace.
      boost::trim(pr);

      // Are we done with the object? Look for the closing brace.
      if (pr[0] == '}')
        break;

      // Get the unescaped name.
      JSONString *namestr;
      ParseString(pr, &namestr);
      std::string name = namestr->str_;
      boost::trim(name);
      delete namestr;

      // Make sure we are at the ":" character.
      boost::trim(pr);
      if (pr[0] != ':')
        throw std::runtime_error("bogus object; name string is not "
                                 "followed by a ':' character.");
      pr.erase(0, 1);
      boost::trim(pr);

      // Make sure the name is not the empty string.
      if (name == "")
        throw std::runtime_error("bogus object; name is empty.");

      // Parse the value.
      JSONValue *val = NULL;
      RecursiveParse(pr, &val);

      // Add the name/val pair.
      if (!((*object)->Put(name, val)) )
        throw std::runtime_error("couldn't object.Put() name/val");

      // Do we have another name/val pair?
      boost::trim(pr);
      if (pr[0] == ',') {
        // Yes.  Strip the comma, loop around and parse it.
        pr.erase(0, 1);
        continue;
      }

      // Nope.  We must be approaching the '}' char.
      break;
    }

    // Skip the '}' char.
    boost::trim(pr);
    if (pr[0] != '}')
      throw std::runtime_error("Bogus object didn't end in '}' "
                               "character -- " + pr);
    pr.erase(0, 1);
  } catch(...) {
    if (*object != NULL)
      delete *object;
    *object = NULL;
    throw;
  }
}

void ParseArray(std::string &pr, JSONArray **array) {
  // STEP 1.  Implement ParseArray.
}

void ParseString(std::string &pr, JSONString **string) {
  // Skip the leading '"' char on the string.
  pr.erase(0, 1);

  // Look ahead, examining each '"' until we find a non-escaped one.
  size_t len = pr.length();
  bool in_escape = false;
  size_t i;
  for (i = 0; i < len; i++) {
    if (in_escape) {
      in_escape = false;
      continue;
    }
    if (pr[i] == '\\') {
      in_escape = true;
      continue;
    }

    // Not in escape, so a quote here is the end of the string.
    if (pr[i] == '\"')
      break;
  }
  if (i == len)
    throw std::runtime_error("Couldn't find end of string in ParseString()");

  // Copy out the string, truncate it (and the trailing ") from pr.
  std::string str = pr.substr(0, i);
  pr.erase(0, i+1);

  // Unescape the string, then create the JSONString object with it.
  std::string unescaped =
    JSONUnescapeString((unsigned char *) str.c_str());
  *string = new JSONString(unescaped.c_str());
  if (*string == NULL)
    throw std::runtime_error("could not allocate new JSONString "
                             "in ParseString");
}

void ParseNumber(std::string &pr, JSONNumber **number) {
  // Use sscanf to parse the C-formatted double.
  double num;
  int charsconsumed;
  int res = sscanf(pr.c_str(), "%lf%n", &num, &charsconsumed);
  if (res != 1)
    throw std::runtime_error("couldn't scanf the float number in ParseNumber");

  // Figure out how many characters to move past the number, and do it.
  // The end of the number is either whitespace, or the end of an object/array.
  size_t offset = pr.find_first_of(" \t\n,]}");
  if (offset == std::string::npos)
    throw std::runtime_error("couldn't parse number in ParseNumber");
  pr.erase(0, offset);
  boost::trim(pr);

  // Allocate/return the new JSONNumber.
  *number = new JSONNumber(num);
  if (*number == NULL)
    throw std::runtime_error("allocating JSONNumber_new failed in ParseNumber");
}

void ParseTrue(std::string &pr, JSONBoolean **t) {
  if (pr.substr(0, 4) == "true") {
    *t = new JSONBoolean(true);
    if (*t == NULL)
      throw std::runtime_error("couldn't allocate new bool true");
    pr.erase(0, 4);
    boost::trim(pr);
    return;
  }
  throw std::runtime_error("couldn't parse true");
}

void ParseFalse(std::string &pr, JSONBoolean **f) {
  if (pr.substr(0, 5) == "false") {
    *f = new JSONBoolean(false);
    if (*f == NULL)
      throw std::runtime_error("couldn't allocate new bool false");
    pr.erase(0, 5);
    boost::trim(pr);
    return;
  }
  throw std::runtime_error("couldn't parse false");
}

void ParseNull(std::string &pr, JSONNull **n) {
  if (pr.substr(0, 4) == "null") {
    *n = new JSONNull();
    if (*n == NULL)
      throw std::runtime_error("couldn't allocate new null");
    pr.erase(0, 4);
    boost::trim(pr);
    return;
  }
  throw std::runtime_error("couldn't parse null");
}
