/*
 * Copyright 2012 Steven Gribble
 *
 *  This file is part of the UW CSE333 project sequence (333proj12).
 *
 *  333proj12 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  333proj12 is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj12.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/algorithm/string/replace.hpp>
#include <boost/algorithm/string.hpp>

#include <ctype.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <string>

#include "./Utils.h"

// Utility function for URIEncode
bool IsOrdinaryChar(char c);

std::string JSONEscapeString(unsigned char *unescaped) {
  uint64_t ue_len = strlen(reinterpret_cast<char *>(unescaped));
  std::string retbuf;

  // Loop through the characters of "unescaped" looking for characters
  // that need to be slash-escaped.
  for (uint64_t i = 0; i < ue_len; i++) {
    if (unescaped[i] < 0x7F) {
      switch (unescaped[i]) {
      case '\"':
        retbuf += '\\';
        retbuf += '\"';
        break;
      case '\\':
        retbuf += '\\';
        retbuf += '\\';
        break;
      case '/':
        retbuf += '\\';
        retbuf += '/';
        break;
      case '\b':
        retbuf += '\\';
        retbuf += 'b';
        break;
      case '\f':
        retbuf += '\\';
        retbuf += 'f';
        break;
      case '\n':
        retbuf += '\\';
        retbuf += 'n';
        break;
      case '\r':
        retbuf += '\\';
        retbuf += 'r';
        break;
      case '\t':
        retbuf += '\\';
        retbuf += 't';
        break;
      default:
        retbuf += unescaped[i];
        break;
      }
    } else {
      // We found a character above ASCII 127, so need convert to
      // slashU encoding (e.g., \u32FA).
      char buf[7] = { 0 };
      int res = UTF8ToSlashUEncoded(&(unescaped[i]), buf);
      retbuf += buf;
      i += res-1;  // rely on for loop to advance the other +1
    }
  }
  return retbuf;
}

std::string JSONUnescapeString(unsigned char *escaped) {
  uint64_t e_len = strlen(reinterpret_cast<char *>(escaped));
  std::string unescaped;

  // Loop through the string, looking for '\\' characters; if found,
  // convert the escaped sequence to an unescaped one.
  for (uint64_t i = 0U; i < e_len; i++) {
    // if a regular character, no need to unescape
    if (escaped[i] != '\\') {
      unescaped += escaped[i];
      continue;
    }

    // is an escape char; do some sanity checking to make sure the full
    // escape sequence is present.
    if (i == e_len-1) {
      // escape char, but no following escaped char!
      fprintf(stderr, "incomplete escaped string '%s'\n", escaped);
      return "";
    }
    if (escaped[i+1] == 'u') {
      if (i+2 >= (e_len-3)) {
        fprintf(stderr, "incomplete slashU-escaped string '%s'\n", escaped);
        return "";
      }
    }

    // Figure out what char(s) was/were escaped, and substitute them in
    // for the escape sequence.
    unsigned char buf[4];
    switch (escaped[i+1]) {
    case '\"':
      unescaped += '\"';
      i+=1;
      break;
    case '\\':
      unescaped += '\\';
      i+=1;
      break;
    case '/':
      unescaped += '/';
      i+=1;
      break;
    case 'b':
      unescaped += '\b';
      i+=1;
      break;
    case 'f':
      unescaped += '\f';
      i+=1;
      break;
    case 'n':
      unescaped += '\n';
      i+=1;
      break;
    case 'r':
      unescaped += '\r';
      i+=1;
      break;
    case 't':
      unescaped += '\t';
      i+=1;
      break;
    case 'u':
      // a slashU is always 6 characters long, so convert then advance
      // i by 5 (and rely on loop to do the final advance).
      SlashUEncodedToUTF8(reinterpret_cast<char *>(&(escaped[i])), buf);
      unescaped += reinterpret_cast<char *>(buf);
      i+=5;
      break;
    default:
      fprintf(stderr, "bogus escape: %c\n", escaped[i+1]);
      return "";
    }
  }
  return unescaped;
}

int UTF8ToSlashUEncoded(unsigned char *bytes, char buf[7]) {
  if (*bytes < 0x80) {
    // don't need UTF encoding for this; is an error!
    // fprintf(stderr, "asked to slash-U encode a code point that"
    //         " is less than U+007F.  No need; error, but doing anyway\n");
    sprintf(buf, "\\u%02hhX%02hhX", 0, bytes[0]);
    return 1;
  } else if (*bytes < 0xE0) {
    // need to convert to UTF-16 from UTF-8 2-byte.  UTF-8 2-byte is:
    // 110xxxxx 10xxxxxx   (i.e., 10 bits worth)
    //   00011100  --> 1C
    //   00000011  --> 03
    //   00111111  --> 3F
    unsigned char hb, lb;
    hb = (bytes[0] & 0x1C) >> 2;
    lb = ((bytes[0] & 0x03) << 6) | (bytes[1] & 0x3F);
    sprintf(buf, "\\u%02hhX%02hhX", hb, lb);
    return 2;
  } else if (*bytes < 0xF0) {
    // need to convert to UTF-16 from UTF-8 3-byte.  UTF-8 3-byte is:
    // 1110xxxx 10xxxxxx 10xxxxxx  (i.e., 16 bits worth)
    //   00001111 --> 0F
    //   00111100 --> 3C
    //   00000011 --> 03
    //   00111111 --> 3F
    unsigned char hb, lb;
    hb = ((bytes[0] & 0x0F) << 4) | ((bytes[1] & 0x3C) >> 2);
    lb = ((bytes[1] & 0x03) << 6) | (bytes[2] & 0x3F);
    sprintf(buf, "\\u%02hhX%02hhX", hb, lb);
    return 3;
  } else {
    // more than UTF-16; is an error!
    fprintf(stderr, "asked to slash-U encode a code point that"
            " is greater than 0xF000; cannot, can only do UTF-16.\n");
    exit(EXIT_FAILURE);
  }
  return 0;
}

void SlashUEncodedToUTF8(char *bytes, unsigned char buf[4]) {
  // Scan in the bytes from the \uXXYY format;  XX --> hb, YY --> lb
  unsigned char hb, lb;
  unsigned int res = sscanf(bytes, "\\u%02hhX%02hhX", &hb, &lb);
  if (res != 2) {
    fprintf(stderr, "cannot convert slashu '%s' -- %d\n", bytes, res);
    // exit(EXIT_FAILURE);
  }

  // Convert hb, lb into a 16 bit unsigned
  uint16_t widechar = (((uint16_t) hb) << 8) | ((uint16_t) lb);

  // Figure out which encoding to do.
  if (widechar <= 0x007F) {
    // straight shot 1 byte
    buf[0] = lb;
    buf[1] = '\0';
  } else if (widechar <= 0x07FF) {
    // convert to 2-byte UTF-8 110xxxyy 10yyyyyy
    unsigned char b1, b2;
    b1 = 0xC0 | ((hb << 2) & 0x1C) | ((lb >> 6) & 0x03);
    b2 = 0x80 | (lb & 0x3F);
    buf[0] = b1;
    buf[1] = b2;
    buf[2] = '\0';
  } else {
    // convert to 3-byte UTF-8 1110xxxx 10xxxxyy 10yyyyyy
    unsigned char b1, b2, b3;
    b1 = 0xE0 | ((hb >> 4) & 0x0F);
    b2 = 0x80 | ((hb << 2) & 0x3C) | ((lb >> 6) & 0x03);
    b3 = 0x80 | (lb & 0x3F);
    buf[0] = b1;
    buf[1] = b2;
    buf[2] = b3;
    buf[3] = '\0';
  }
}

std::string URIEncode(const std::string &src) {
  // adapted from http://www.codeguru.com/cpp/cpp/algorithms/strings/
  // article.php/c12759/URI-Encoding-and-Decoding.htm
  std::string retstr;
  int len = src.length();
  for (int i = 0; i < len; i++) {
    if (IsOrdinaryChar(src[i])) {
      retstr += src[i];
    } else if (src[i] == ' ') {
      retstr += "+";
    } else {
      char tmp[6];
      sprintf(tmp, "%%%x", src[i]);
      retstr += tmp;
    }
  }
  return retstr;
}

bool IsOrdinaryChar(char c) {
  char ch = tolower(c);
  if (ch == 'a' || ch == 'b' || ch == 'c' || ch == 'd' || ch == 'e' ||
      ch == 'f' || ch == 'g' || ch == 'h' || ch == 'i' || ch == 'j' ||
      ch == 'k' || ch == 'l' || ch == 'm' || ch == 'n' || ch == 'o' ||
      ch == 'p' || ch == 'q' || ch == 'r' || ch == 's' || ch == 't' ||
      ch == 'u' || ch == 'v' || ch == 'w' || ch == 'x' || ch == 'y' ||
      ch == 'z' || ch == '0' || ch == '1' || ch == '2' || ch == '3' ||
      ch == '4' || ch == '5' || ch == '6' || ch == '7' || ch == '8' ||
      ch == '9') {
    return true;
  }
  return false;
}

std::string EscapeHTMLInPlace(const std::string &from) {
  std::string ret = from;
  boost::replace_all(ret, "&",  "&amp;");
  boost::replace_all(ret, "\"", "&quot;");
  boost::replace_all(ret, "\'", "&apos;");
  boost::replace_all(ret, "<",  "&lt;");
  boost::replace_all(ret, ">",  "&gt;");
  return ret;
}
